# Copyright (c) 2019-2020, CPB Netherlands Bureau for Economic Policy Analysis
# Copyright (c) 2019-2020, Andrey Dubovik <andrei@dubovik.eu>

"""Routines for downloading directory listings."""

# Load global packages
import re
from urllib.parse import urlparse, urljoin
from lxml import etree
import random
import threading
import sqlite3

# Load local packages
from . import settings
from .misc import iserror, silent, parallel
from .network import fetch, fetch_curl


def listing_ftp(url):
    """Download an FTP directory listing."""
    files = fetch_curl(url, "-l").decode().splitlines()
    prog = re.compile(r".+\.(rdf|redif)$", flags=re.I)
    files = [f for f in files if prog.match(f)]
    return [url + f for f in files]


def listing_http(url):
    """Download an HTTP directory listing."""
    content, _ = fetch(url)
    html = etree.HTML(content)
    files = html.xpath("//a/@href")
    prog = re.compile(r".+\.(rdf|redif)$", flags=re.I)
    files = [f for f in files if prog.match(f)]
    return [urljoin(url, f) for f in files]


@silent
def listing(url):
    """Get a list of ReDIF files for a given series."""
    scheme = urlparse(url)[0]
    if scheme == "ftp":
        files = listing_ftp(url)
    elif scheme in ["http", "https"]:
        files = listing_http(url)
    else:
        raise RuntimeError("Unknown scheme {}".format(scheme))
    if len(files) == 0:
        raise RuntimeError("Empty listing")
    return files


def update_listings_1(conn, lock, url):
    """Update remote listings for a single series."""
    files = listing(url)
    with lock:
        c = conn.cursor()
        if iserror(files):
            sql = "UPDATE remotes SET status = 2, error = ? WHERE url = ?"
            c.execute(sql, (str(files), url))
        else:
            files = [(f, url) for f in files]
            sql = "UPDATE remotes SET status = 0, error = NULL WHERE url = ?"
            c.execute(sql, (url,))
            sql = "REPLACE INTO listings (url, remote) VALUES (?, ?)"
            c.executemany(sql, files)
        c.close()
    return not iserror(files)


def update_listings(conn, lock, status=1):
    """Update remote listings for all series."""
    c = conn.cursor()
    c.execute("SELECT url FROM remotes WHERE status = ?", (status,))
    urls = [r[0] for r in c.fetchall()]
    urls = random.sample(urls, k=len(urls))  # to redistribute load
    c.close()
    print("Updating remote listings...")

    def worker(u):
        return update_listings_1(conn, lock, u)

    status = parallel(worker, urls, settings.no_threads_www)
    print(f"{sum(status)} out of {len(urls)} records updated successfully")


def update(handle: str = ""):
    """Update remote listings (wrapper)."""
    conn = sqlite3.connect(settings.database, check_same_thread=False)
    lock = threading.Lock()
    try:
        if len(handle):
            c = conn.cursor()
            c.execute("select url from series where lower(handle)=lower(?);", (handle,))
            u = c.fetchone()[0]
            update_listings_1(conn, lock, u)
        else:
            update_listings(conn, lock)
    except BaseException:
        conn.rollback()
        raise
    else:
        conn.commit()
    conn.close()



# Copyright (c) 2019-2020, CPB Netherlands Bureau for Economic Policy Analysis
# Copyright (c) 2019-2020, Andrey Dubovik <andrei@dubovik.eu>

"""A patch to add missing journal names from series names.

This patch is not yet integrated into the main code.
"""

# Load local packages
from . import settings
from .misc import dbconnection, collect
from . import redif
from .repec import ftp_get


@dbconnection(settings.database)
def fetch_handles(conn):
    """Return a set of handle prefixes that miss journal names."""
    sql = "SELECT handle FROM papers" " WHERE template = 'article' AND journal IS NULL"
    handles = conn.execute(sql).fetchall()
    return set(":".join(h[0].split(":")[:3]) for h in handles)


@dbconnection(settings.database)
def fetch_files(conn, handles):
    """Return a set of files containing given handles."""
    sql = "SELECT file FROM series WHERE handle = ?"
    c = conn.cursor()
    files = set()
    for handle in handles:
        for (file,) in c.execute(sql, (handle,)):
            files.add(file)
    return files


def collect_names(files):
    """Download files and collect handle -> name associations."""
    handles = {}
    for i, file in enumerate(files):
        print(f"[{i+1}/{len(files)}] {file}...")
        try:
            rdf = redif.load(redif.decode(ftp_get(settings.repec_ftp + file)))
            for record in rdf:
                record = collect(record)
                if "name" in record:
                    # Account for inconsistent capitalization across records
                    handle = record["handle"][0].lower()
                    newname = record["name"][0]
                    oldname = handles.setdefault(handle, newname)
                    if newname != oldname:
                        print(
                            f'Conflicting names: "{oldname}" vs. "{newname}"'
                            f" in {handle}"
                        )
        except Exception:
            print(f"Skipping {file} due to errors")
    return handles


def merge(handles, names):
    """Intersect handles and names correcting for capitalization."""
    return {h: names[h.lower()] for h in handles if h.lower() in names}


@dbconnection(settings.database)
def ensure_indices(conn):
    """Create necessary indices if missing."""
    sql = """
        CREATE INDEX IF NOT EXISTS papers_tejoha
        ON papers (template, journal, handle);
    """
    conn.execute(sql)


@dbconnection(settings.database)
def update_names(conn, names):
    """Update journal names from series names."""
    sql = """
        UPDATE papers SET journal = ?
        WHERE template = 'article' AND journal IS NULL AND handle GLOB '{}*'
    """
    c = conn.cursor()
    for handle, name in names.items():
        c.execute(sql.format(handle), (name,))


def main():
    """Run the whole update."""
    ensure_indices()
    handles = fetch_handles()
    files = fetch_files(handles=handles)
    names = collect_names(files)
    names = merge(handles, names)
    update_names(names=names)


if __name__ == "__main__":
    main()



# Copyright (c) 2019-2020, CPB Netherlands Bureau for Economic Policy Analysis
# Copyright (c) 2019-2020, Andrey Dubovik <andrei@dubovik.eu>

"""Command line interface."""

# Load standard packages
import argparse

# Load local packages
from . import settings
from . import database
from . import repec
from . import remotes
from . import papers


def init(args):
    """Initialize the database."""
    settings.database = args.database
    database.prepare(settings.database)


def update(args):
    """Run full database update."""
    settings.database = args.database
    settings.timeout = args.timeout
    settings.batch_size = args.batchsize
    settings.no_threads_repec = args.threads_repec
    settings.no_threads_www = args.threads_www
    settings.verbosity = max(settings.verbosity - args.quiet, 0)

    if not args.repec and not args.listings and not args.papers:
        args.repec = args.listings = args.papers = True
    if len(args.listings_of_handle):
        args.repec = args.listings = args.papers = False

    database.check_version()  # Abort on incompatible versions

    if args.repec:
        repec.update()
    if args.listings_of_handle:
        remotes.update(args.listings_of_handle)
    if args.listings:
        remotes.update()
    if args.papers:
        papers.update()


def main():
    parser = argparse.ArgumentParser(description="RePEc Database Manager")
    commands = parser.add_subparsers(title="Subcommands", required=True)

    # Initialize subcommand

    p_init = commands.add_parser(
        "init",
        help="Initialize the database",
        description="Initialize the database, download JEL codes",
    )
    p_init.set_defaults(func=init)
    p_init.add_argument(
        "--database",
        default=settings.database,
        help=f"SQLite database location (default: {settings.database})",
    )

    # Update subcommand

    p_update = commands.add_parser(
        "update",
        help="Run full database update",
        description=(
            "Run database update. If none of --repec, --listings, or --papers"
            " is given, full update is performed. Otherwise, the selected"
            " parts are updated."
        ),
    )
    p_update.set_defaults(func=update)
    p_update.add_argument(
        "--database",
        default=settings.database,
        help=f"SQLite database location (default: {settings.database})",
    )
    p_update.add_argument(
        "--repec",
        action="store_true",
        help="Download RePEc website data",
    )
    p_update.add_argument(
        "--listings",
        action="store_true",
        help="Download website listings",
    )
    p_update.add_argument(
        "--listings-of-handle",
        type=str,
        default="",
        help="Download website listings for a given handle",
    )
    p_update.add_argument(
        "--papers",
        action="store_true",
        help="Download papers",
    )
    p_update.add_argument(
        "--timeout",
        type=int,
        default=settings.timeout,
        help=("Timeout for individual requests, sec" f" (default: {settings.timeout})"),
    )
    p_update.add_argument(
        "--batchsize",
        type=int,
        default=settings.batch_size,
        metavar="SIZE",
        help=(
            "Number of papers to commit in a single transaction"
            f" (default: {settings.batch_size:,})"
        ),
    )
    p_update.add_argument(
        "--threads-repec",
        type=int,
        default=settings.no_threads_repec,
        metavar="N",
        help=(
            "Number of threads when downloading from RePEc website"
            f" (default: {settings.no_threads_repec})"
        ),
    )
    p_update.add_argument(
        "--threads-www",
        type=int,
        default=settings.no_threads_repec,
        metavar="N",
        help=(
            "General number of threads used for downloading"
            f" (default: {settings.no_threads_www})"
        ),
    )
    p_update.add_argument(
        "-q",
        "--quiet",
        action="count",
        default=0,
        help="Decrease verbosity level (can be given multiple times)",
    )

    # Process command line arguments and dispatch on subcommand
    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()



# Copyright (c) 2019-2020, CPB Netherlands Bureau for Economic Policy Analysis
# Copyright (c) 2019-2020, Andrey Dubovik <andrei@dubovik.eu>

"""Miscellaneous routines."""

# Load global packages
from concurrent.futures import ThreadPoolExecutor
from functools import wraps
import sqlite3

# Load local packages
from . import settings


def iserror(err):
    """Check if err is an exception."""
    return isinstance(err, Exception)


def silent(func):
    """Catch errors and return them regularly."""

    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as err:
            return err

    return wrapper


def meter(it, it_len):
    """Print progress information."""
    for i, el in enumerate(it):
        if settings.verbosity == 3:
            print("[{}/{}]".format(i + 1, it_len), end="\r")
        yield el


def parallel(func, lst, threads):
    """Apply a function to a list in parallel."""
    with ThreadPoolExecutor(max_workers=threads) as executor:
        return list(meter(executor.map(func, lst), len(lst)))


def collect(lst):
    """Collect a list of key, value pairs into a dictionary."""
    dct = {}
    for k, v in lst:
        vv = dct.setdefault(k, [])
        vv.append(v)
    return dct


def dbconnection(database):
    """Establish a database connection."""

    def decorator(func):
        @wraps(func)
        def wrapper(**kwargs):
            conn = sqlite3.connect(database)
            try:
                rslt = func(conn=conn, **kwargs)
                conn.commit()
            except BaseException:
                conn.rollback()
                raise
            finally:
                conn.close()
            return rslt

        return wrapper

    return decorator



# Copyright (c) 2021, Andrey Dubovik <andrei@dubovik.eu>

"""Convenience wrappers around network functions."""

# Load global packages
import requests
from requests.packages import urllib3
import subprocess

# Load local packages
from . import settings


def fetch(url):
    """Fetch an URL using requests."""
    try:
        headers = {"User-Agent": settings.user_agent}
        response = requests.get(url, timeout=settings.timeout, headers=headers)
    except requests.exceptions.ConnectionError as err:
        if type(err.args[0]) == urllib3.exceptions.MaxRetryError:
            err.args = ("Max retries exceeded",)  # Simplify error message
            raise
        else:
            raise
    if response.status_code != 200:
        raise RuntimeError("HTTP Error {}".format(response.status_code))
    return response.content, response.encoding


def fetch_curl(url, options=""):
    """Fetch an URL using curl."""
    cmd = ["curl", "-sm {}".format(settings.timeout), options, url]
    rslt = subprocess.run(cmd, stdout=subprocess.PIPE)
    if rslt.returncode != 0:
        raise RuntimeError("CURL Error {}".format(rslt.returncode))
    return rslt.stdout



# Copyright (c) 2019-2020, CPB Netherlands Bureau for Economic Policy Analysis
# Copyright (c) 2019-2020, Andrey Dubovik <andrei@dubovik.eu>

"""Database intialization routines."""

# Load global packages
import sqlite3
import os
from lxml import etree
from html import unescape

# Load local packages
from . import settings
from .network import fetch

# Define constants
DBVERSION = "9"

SQL = f"""
    CREATE TABLE repec (
        file text PRIMARY KEY,
        type text,
        ftpdate text,
        status integer DEFAULT 1,
        error text,
        replaced_at text DEFAULT CURRENT_TIMESTAMP
    );
    CREATE TABLE series (
        file text,
        handle text,
        type text,
        url text,
        status integer DEFAULT 1,
        error text,
        replaced_at text DEFAULT CURRENT_TIMESTAMP,
        PRIMARY KEY (file, handle)
    );
    CREATE TABLE remotes (
        url text PRIMARY KEY,
        file text,
        handle text,
        status integer DEFAULT 1,
        error text,
        replaced_at text DEFAULT CURRENT_TIMESTAMP
    );
    CREATE TABLE listings (
        url text PRIMARY KEY,
        remote text,
        status integer DEFAULT 1,
        error text,
        replaced_at text DEFAULT CURRENT_TIMESTAMP
    );
    CREATE TABLE papers (
        pid integer PRIMARY KEY,
        handle text UNIQUE,
        url text,
        template text,
        language text,
        title text,
        abstract text,
        journal text,
        year integer,
        volume integer,
        issue integer,
        pages text,
        redif blob,
        replaced_at text DEFAULT CURRENT_TIMESTAMP
    );
    CREATE TABLE authors (
        pid integer REFERENCES papers ON DELETE CASCADE,
        name text,
        email text
    );
    CREATE INDEX authors_pid ON authors (pid);
    CREATE TABLE jel (
        code text PRIMARY KEY,
        parent text,
        description text
    );
    CREATE TABLE papers_jel (
        pid integer REFERENCES papers ON DELETE CASCADE,
        code text REFERENCES jel,
        PRIMARY KEY (pid, code)
    );
    CREATE INDEX papers_jel_code ON papers_jel (code);
    CREATE TABLE meta (
        parameter text PRIMARY KEY,
        value text
    );
    INSERT INTO meta VALUES
        ('version', {DBVERSION});
"""


def jcode(item):
    """Get JEL code."""
    e = item.xpath("code/text()")
    return e[0] if e else None


def jdesc(item):
    """Get JEL description."""
    return unescape(item.xpath("description/text()")[0])


def import_level(c, element):
    """Recursively import JEL hierarchy."""
    sql = "INSERT INTO jel (code, parent, description) VALUES (?, ?, ?)"
    parent = jcode(element)
    items = element.xpath("classification")
    rows = [(jcode(i), parent, jdesc(i)) for i in items]
    c.executemany(sql, rows)
    for item in items:
        import_level(c, item)


def populate_jel(conn):
    """Download and save official JEL classification."""
    content, _ = fetch(settings.jel)
    xml = etree.fromstring(content)
    with conn:
        c = conn.cursor()
        import_level(c, xml)
        c.close()


def prepare(path):
    """Prepare a new SQLite database."""
    if os.path.exists(path):
        raise RuntimeError("Database already exists")
    conn = sqlite3.connect(path)
    conn.executescript(SQL)
    populate_jel(conn)


def check_version():
    """Verify database version."""
    sql = "SELECT value FROM meta WHERE parameter = 'version'"
    conn = sqlite3.connect(settings.database)
    (version,) = conn.execute(sql).fetchone()
    if version != DBVERSION:
        raise RuntimeError("Incompatible database version")



# Copyright (c) 2019-2020, CPB Netherlands Bureau for Economic Policy Analysis
# Copyright (c) 2019-2020, Andrey Dubovik <andrei@dubovik.eu>

"""Routines for downloading data from the RePEc website."""

# Load global packages
import re
import threading
import sqlite3
from datetime import datetime

# Load local packages
from . import settings
from . import redif
from .misc import iserror, silent, parallel, collect
from .network import fetch_curl

# RePEc FTP is broken; using curl as a workaround


def ftp_datetime(month, day, tory):
    """Decode ftp modification datetime."""
    if tory.find(":") != -1:
        time, year = tory, str(datetime.now().year)
    else:
        time, year = "00:00", tory
    dt = " ".join((year, month, day, time))
    return datetime.strptime(dt, "%Y %b %d %H:%M")


def ftp_ls(url):
    """Get file listing (with modification dates)."""
    files = fetch_curl(url).decode().splitlines()
    prog = re.compile(r"\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+(\S+)\s+(\S+)\s+(\S+)\s+(.+)")
    files = [prog.match(f).groups() for f in files]
    return [(f[3], ftp_datetime(*f[:3])) for f in files]


def update_repec(conn):
    """Update the list of available archives."""
    files = ftp_ls(settings.repec_ftp)
    prog = re.compile(r"(...(arch|seri)\.(rdf|redif))$", flags=re.I)
    files = [(prog.match(f), d) for f, d in files]
    files = [(*m.groups(), d) for m, d in files if m]
    files = [(n, t.lower(), d.isoformat(sep=" ")) for n, t, e, d in files]
    c = conn.cursor()
    sql = "REPLACE INTO repec (file, type, ftpdate) VALUES (?, ?, ?)"
    c.executemany(sql, files)
    c.close()


@silent
def load(file, cat):
    """Load a series or an archive file."""
    rdf = redif.load(redif.decode(fetch_curl(settings.repec_ftp + file)))

    def key(r):
        r = dict(r)
        url = r["url"].rstrip("/") + "/" if "url" in r else None
        return (file, cat, r["handle"], url)

    return [key(r) for r in rdf]


def update_series_1(conn, lock, file, cat):
    """Update the database for a series or an archive file."""
    r = load(file, cat)
    with lock:
        c = conn.cursor()
        if iserror(r):
            sql = "UPDATE repec SET status = 2, error = ? WHERE file = ?"
            c.execute(sql, (str(r), file))
        else:
            sql = "UPDATE repec SET status = 0, error = NULL WHERE file = ?"
            c.execute(sql, (file,))
            sql = "REPLACE INTO series (file, type, handle, url)" " VALUES (?, ?, ?, ?)"
            c.executemany(sql, r)
        c.close()
    return not iserror(r)


def update_series(conn, lock, status=1):
    """Update all archive and series files in the database."""
    c = conn.cursor()
    c.execute("SELECT file, type FROM repec WHERE status = ?", (status,))
    files = c.fetchall()
    c.close()
    print("Updating archive and series files...")

    def worker(el):
        return update_series_1(conn, lock, *el)

    status = parallel(worker, files, threads=settings.no_threads_repec)
    print(f"{sum(status)} out of {len(files)} records updated successfully")


def update_remotes(conn, status=1):
    """Update the list of remotes in the database."""
    c = conn.cursor()

    # Archives
    sql = "SELECT handle, url FROM series, repec USING (file)"
    sql += ' WHERE series.type = "arch" ORDER BY ftpdate DESC'
    c.execute(sql)
    archives = [(h.lower(), u) for h, u in c.fetchall()]
    archives = {k: v[0] for k, v in collect(archives).items()}
    c.execute('UPDATE series SET status = 0 WHERE type = "arch"')

    # Series
    def lookup(s):
        a, sep, f = s[1].rpartition(":")
        a = archives.get(a.lower())
        if a:
            return (a + f + "/", 0, None, *s)
        else:
            return (None, 2, "Archive not found", *s)

    sql = 'SELECT file, handle FROM series WHERE type = "seri" AND status = ?'
    c.execute(sql, (status,))
    series = [lookup(s) for s in c.fetchall()]
    sql = (
        "UPDATE series SET url = ?, status = ?, error = ?"
        " WHERE file = ? AND handle = ?"
    )
    c.executemany(sql, series)

    # Remotes
    remotes = list(set((s[0], s[-2], s[-1]) for s in series if s[0]))
    c.executemany("REPLACE INTO remotes (url,file,handle) VALUES (?,?,?)", remotes)
    c.close()


def update():
    """Update the database on the basis of RePEc information."""
    conn = sqlite3.connect(settings.database, check_same_thread=False)
    lock = threading.Lock()
    try:
        update_repec(conn)
        update_series(conn, lock)
        update_remotes(conn)
    except BaseException:
        conn.rollback()
        raise
    else:
        conn.commit()
    conn.close()



# Copyright (c) 2019-2020, CPB Netherlands Bureau for Economic Policy Analysis
# Copyright (c) 2019-2020, Andrey Dubovik <andrei@dubovik.eu>

"""Routines for downloading and destructuring papers."""

# Load global packages
from urllib.parse import urlparse
import re
import sqlite3
import json
import zlib
import threading
import random
import math
import cld2
from collections import defaultdict

# Load local packages
from . import settings
from . import redif
from .misc import iserror, silent, parallel
from .sanitize import sanitize, sanitize_email
from .network import fetch, fetch_curl


def ttype(record):
    """Get template type."""
    tt = next(v for k, v in record if k == "template-type")
    return re.match(r"redif-(\S*)", tt.lower()).group(1)


@silent
def load(url):
    """Download ReDIF papers."""
    scheme = urlparse(url)[0]
    if scheme == "ftp":
        papers = redif.decode(fetch_curl(url))
    elif scheme in ["http", "https"]:
        content, encoding = fetch(url)
        papers = redif.decode(content, hint=[encoding])
    else:
        raise RuntimeError("Unknown scheme {}".format(scheme))
    papers = redif.load(papers)
    for p in papers:
        fields = set(k for k, v in p)
        for f in ["handle", "template-type"]:
            if f not in fields:
                raise RuntimeError("{} is missing".format(f))
    if len(papers) == 0:
        raise RuntimeError("Empty series")
    return papers


def filterjel(jel, alljel):
    """Verify a JEL code against the official list."""
    if jel in alljel:
        return jel
    elif jel[:2] in alljel:
        return jel[:2]


def parsejel(jel, alljel):
    """Parse JEL using ad-hoc rules."""
    jel = re.sub("([A-Z])[-., ]+([0-9])", r"\1\2", jel)
    jel = re.split("[^A-Z0-9]+", jel.upper())
    jel = [c[:3] for c in jel if re.match("[A-Z][0-9]+$", c)]
    jel = [filterjel(c, alljel) for c in jel]
    jel = sorted(set(c for c in jel if c))
    # Do not include JEL for papers that blindly follow the online example
    if jel == ["R00", "Z0"]:
        return []
    return jel


def parse_template(template):
    """Parse broken template specification."""
    m = re.search("edif-([a-z]+)", template, flags=re.I)
    return m.group(1).lower() if m else None


def parse_year(date):
    """Parse broken date specification."""
    m = re.search("(?<![0-9])[0-9]{4}", date)
    y = m.group(0) if m else None
    return int(y) if y and y != "0000" else None


def get_year(paper):
    """Get most relevant available year."""
    for f in ["year", "creation-date", "revision-date"]:
        years = [parse_year(d) for d in paper.get(f, [])]
        years = [y for y in years if y]
        if years:
            return min(years)
    return None


def detect_language(text):
    """Detect language using CLD2 library."""
    try:  # todo: figure out what's causing an occasional error
        _, _, details = cld2.detect(text)
        lang = details[0].language_code
    except ValueError:
        lang = "un"
    return lang if lang != "un" else None


def lang_and(*text, default=None):
    """Determine common language."""
    lang = set(detect_language(t) for t in text if t)
    if len(lang) == 1:
        lang = lang.pop()
        if lang:
            return lang
    return default


def replace_paper(c, paper, url, alljel):
    """Update a single paper record."""
    blob = json.dumps(paper, ensure_ascii=False).encode(encoding="utf-8")
    paper = redif.collect(paper)
    r = {}
    r["url"] = url
    r["handle"] = paper["handle"][0]
    r["template"] = parse_template(paper["template-type"][0])
    for f in ["title", "abstract", "journal", "volume", "issue", "pages"]:
        r[f] = paper.get(f, [None])[0]
    for f in ["title", "abstract", "journal"]:
        r[f] = sanitize(r[f])
    r["language"] = paper.get("language", ["none"])[0].lower()
    r["language"] = r["language"] if len(r["language"]) == 2 else None
    r["language"] = lang_and(r["title"], r["abstract"], default=r["language"])
    r["year"] = get_year(paper)
    r["redif"] = zlib.compress(blob, level=9)

    sql = "REPLACE INTO papers (" + ", ".join(k for k in r.keys()) + ")"
    sql += " VALUES (" + ", ".join(["?"] * len(r)) + ")"
    c.execute(sql, list(r.values()))
    pid = c.lastrowid

    if "author" in paper:
        authors = [a for a in paper["author"] if type(a) == defaultdict]
        authors = [
            (sanitize(a["name"][0]), sanitize_email(a.get("email", [None])[0]))
            for a in authors
        ]
        authors = [(pid, n, e) for n, e in authors if n]
        sql = "INSERT INTO authors (pid, name, email) VALUES (?, ?, ?)"
        c.executemany(sql, authors)
    if "classification-jel" in paper:
        jel = parsejel(paper["classification-jel"][0], alljel)
        jel = [(pid, c) for c in jel]
        c.executemany("INSERT INTO papers_jel (pid, code) VALUES (?, ?)", jel)


def update_papers_1(conn, lock, url, alljel):
    """Update papers from a single ReDIF document."""
    papers = load(url)
    with lock:
        c = conn.cursor()
        if iserror(papers):
            sql = "UPDATE listings SET status = 2, error = ? WHERE url = ?"
            c.execute(sql, (str(papers), url))
        else:
            sql = "UPDATE listings SET status = 0, error = NULL WHERE url = ?"
            c.execute(sql, (url,))
            for paper in papers:
                replace_paper(c, paper, url, alljel)
        c.close()
    return not iserror(papers)


def update_papers(conn, lock, status=1):
    """Update papers from all ReDIF documents."""
    c = conn.cursor()
    c.execute("SELECT code FROM jel WHERE parent IS NOT NULL")
    alljel = [r[0] for r in c.fetchall()]
    c.execute("SELECT url FROM listings WHERE status = ?", (status,))
    urls = [r[0] for r in c.fetchall()]
    urls = random.sample(urls, k=len(urls))  # to redistribute load
    c.close()

    def worker(u):
        return update_papers_1(conn, lock, u, alljel)

    size = settings.batch_size
    no_batches = math.ceil(len(urls) / size)
    status = 0
    for i in range(no_batches):
        print("Downloading batch {}/{}...".format(i + 1, no_batches))
        batch = urls[i * size : (i + 1) * size]
        bs = sum(parallel(worker, batch, threads=settings.no_threads_www))
        status += bs
        conn.commit()
        print(f"{bs} out of {len(batch)} records updated successfully")

    print(f"All batches: {status} out of {len(urls)} records" " updated successfully")


def update():
    """Update papers from all ReDIF documents (wrapper)."""
    conn = sqlite3.connect(settings.database, check_same_thread=False)
    c = conn.cursor()
    c.execute("PRAGMA foreign_keys = ON")
    c.close()

    lock = threading.Lock()
    try:
        update_papers(conn, lock)
    except BaseException:
        conn.rollback()
        raise
    else:
        conn.commit()
    conn.close()



# Copyright (c) 2019-2020, CPB Netherlands Bureau for Economic Policy Analysis
# Copyright (c) 2019-2020, Andrey Dubovik <andrei@dubovik.eu>

"""Configuration parameters."""

# Load libraries
import os

# Default command line arguments
database = './repec.db'
timeout = 300  # seconds
batch_size = 10000  # records in each commit
no_threads_repec = 32
no_threads_www = 128
verbosity = 3

# Additional configuration
repec_ftp = 'ftp://all.repec.org/RePEc/all/'
jel = 'https://www.aeaweb.org/econlit/classificationTree.xml'
curl = 'curl' if os.name == 'nt' else '/usr/bin/curl'

# Masquerading as Chrome, otherwise some sites refuse connection
user_agent = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64)'
user_agent += ' AppleWebKit/537.36 (KHTML, like Gecko)'
user_agent += ' Chrome/74.0.3729.157 Safari/537.36'



from setuptools import setup, find_packages

setup(
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    entry_points={
        "console_scripts": ["repec=repec.main:main"],
    },
)



# Copyright (c) 2019-2020, CPB Netherlands Bureau for Economic Policy Analysis
# Copyright (c) 2019-2020, Andrey Dubovik <andrei@dubovik.eu>

"""Text and html sanitization routines."""

# Load packages
import re
from lxml.html import tostring, html5parser
import warnings

# Define global settings
BLOCKTAGS = ["div", "p", "br", "li", "h1", "h2", "h3", "h4", "h5", "h6"]

EMAIL = re.compile("['a-z0-9._-]+@[a-z0-9._-]+.[a-z]+")


def remove_cc(text):
    r"""Remove control characters (except for \t, \r, \n)."""
    cc = "[\x00-\x08\x0b\x0c\x0e-\x1f\x7f\x80-\x9f]"
    return re.sub(cc, "", text)


def sanitize_entity(text):
    """Remove whitespace characters from HTML entities."""

    def strip(m):
        return re.sub(r"\s", "", m.group(0))

    return re.sub(r"&#x?[0-9a-f\s]+;", strip, text, flags=re.I)


def ishtml(text):
    """Guess whether text has html in it."""
    text = text.lower()
    for t in BLOCKTAGS:
        if text.find(f"<{t}>") != -1:
            return True
    if text.find("</") != -1 or text.find("/>") != -1:
        return True
    if re.search("&#?x?[0-9a-z]+;", text):
        return True
    return False


def html2text(html):
    """Render html as text, convert line breaks to spaces."""
    if not ishtml(html):
        return re.sub(r"\s+", " ", html.strip())
    parser = html5parser.HTMLParser(namespaceHTMLElements=False)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        html = html5parser.fromstring(html, parser=parser)
    for b in BLOCKTAGS:
        for e in html.xpath(f"//{b}"):
            e.text = " " + e.text if e.text else ""
            if len(e) > 0:
                lc = e[-1]
                lc.tail = (lc.tail if lc.tail else "") + " "
            else:
                e.text = e.text + " "
    text = tostring(html, method="text", encoding="utf-8")
    return re.sub(r"\s+", " ", text.decode().strip())


def isna(token):
    """Check if it's an N/A token."""
    if re.sub(r"\W", "", token).lower() == "na":
        return True
    else:
        return False


def isvalid(token):
    """Check if token contains alpha characters."""
    if re.match(r"^[\W0-9]*$", token):
        return False
    else:
        return True


def sanitize(text):
    """Remove control characters, drop HTML tags, etc."""
    if type(text) != str:
        return text
    text = remove_cc(text)
    text = sanitize_entity(text)
    text = html2text(text)
    if isna(text) or not isvalid(text):
        return None
    return text


def sanitize_email(email):
    """Get a clean email address."""
    email = sanitize(email)
    if type(email) != str:
        return email
    m = EMAIL.search(email.lower())
    return m.group() if m else None



# Copyright (c) 2019-2020, CPB Netherlands Bureau for Economic Policy Analysis
# Copyright (c) 2019-2020, Andrey Dubovik <andrei@dubovik.eu>

"""Routines for decoding and parsing ReDIF documents."""

# Load global packages
import re
from collections import defaultdict

# Define constants
CLUSTERS = set([
    'author-name',
    'editor-name',
    'file-url',
    'provider-name',
    'workplace-name',
])


def decode(rdf, hint=[]):
    """Decode ReDIF document."""
    def decode(encoding):
        rslt = rdf.decode(encoding)
        if rslt.lower().find('template-type') == -1:
            raise RuntimeError('Decoding Error')
        return rslt

    encodings = hint + ['windows-1252', 'utf-8', 'utf-16', 'latin-1']
    if rdf[:3] == b'\xef\xbb\xbf':
        encodings = ['utf-8-sig'] + encodings
    for enc in encodings:
        try:
            return decode(enc)
        except Exception:
            continue
    raise RuntimeError('Decoding Error')


def split(lst, sel):
    """Split a list using a selector function."""
    group = []
    groups = [group]
    for el in lst:
        if sel(el):
            group = []
            groups.append(group)
        group.append(el)
    return groups


def load(rdf):
    """Load ReDIF document."""
    # Repair line endings
    rdf = re.sub('\r(?!\n)', '\r\n', rdf, flags=re.M)

    # Drop comments
    rdf = re.sub('^#.*\n?', '', rdf, flags=re.M)

    # Split fields
    rdf = re.split(r'(^[a-zA-Z0-9\-#]+:\s*)', rdf, flags=re.M)[1:]
    rdf = [line.strip() for line in rdf]
    rdf = [
        (rdf[i].rstrip(':').lower(), rdf[i+1])
        for i in range(0, len(rdf), 2)
    ]

    # Drop empty fields
    rdf = [f for f in rdf if f[1] != '']

    # Split templates
    rdf = split(rdf, lambda x: x[0] == 'template-type')[1:]
    return rdf


def collect(records):
    """Collect ReDIF fields together, group clusters."""
    def helper(path, head, doc, i):
        flag = False
        while i < len(records):
            k, v = records[i]
            if k.startswith(path):
                k = k[len(path):]
                if k in CLUSTERS:
                    # Enter new cluster
                    base, _, key = k.partition('-')
                    cluster = defaultdict(lambda: [])
                    doc[base].append(cluster)
                    i = helper(path + base + '-', key, cluster, i)
                elif flag and k == head:
                    # Exit cluster on a repeated cluster key
                    return i
                else:
                    doc[k].append(v)
                    flag = True
                    i += 1
            else:
                # Exit cluster
                return i
        return i

    doc = defaultdict(lambda: [])
    helper('', '', doc, 0)
    return doc
